@@ -20,10 +20,8 @@ import java.io.PrintWriter;
 import java.io.StringWriter;
 import java.io.Writer;
 import java.lang.management.ManagementFactory;
-import java.util.ArrayList;
+import java.lang.ref.WeakReference;
 import java.util.Iterator;
-import java.util.LinkedList;
-import java.util.List;
 import java.util.TimerTask;
 import java.util.concurrent.atomic.AtomicLong;
 
@@ -68,6 +66,7 @@ public abstract class BaseGenericObjectPool<T> {
     private volatile long maxWaitMillis =
             BaseObjectPoolConfig.DEFAULT_MAX_WAIT_MILLIS;
     private volatile boolean lifo = BaseObjectPoolConfig.DEFAULT_LIFO;
+    private final boolean fairness;
     private volatile boolean testOnCreate =
             BaseObjectPoolConfig.DEFAULT_TEST_ON_CREATE;
     private volatile boolean testOnBorrow =
@@ -94,11 +93,12 @@ public abstract class BaseGenericObjectPool<T> {
     private Evictor evictor = null; // @GuardedBy("evictionLock")
     Iterator<PooledObject<T>> evictionIterator = null; // @GuardedBy("evictionLock")
     /*
-     * Class loader for evictor thread to use since in a J2EE or similar
-     * environment the context class loader for the evictor thread may have
-     * visibility of the correct factory. See POOL-161.
+     * Class loader for evictor thread to use since, in a JavaEE or similar
+     * environment, the context class loader for the evictor thread may not have
+     * visibility of the correct factory. See POOL-161. Uses a weak reference to
+     * avoid potential memory leaks if the Pool is discarded rather than closed.
      */
-    private final ClassLoader factoryClassLoader;
+    private final WeakReference<ClassLoader> factoryClassLoader;
 
 
     // Monitoring (primarily JMX) attributes
@@ -110,12 +110,11 @@ public abstract class BaseGenericObjectPool<T> {
     final AtomicLong destroyedCount = new AtomicLong(0);
     final AtomicLong destroyedByEvictorCount = new AtomicLong(0);
     final AtomicLong destroyedByBorrowValidationCount = new AtomicLong(0);
-    private final LinkedList<Long> activeTimes = new LinkedList<Long>(); // @GuardedBy("activeTimes") - except in initStats()
-    private final LinkedList<Long> idleTimes = new LinkedList<Long>(); // @GuardedBy("activeTimes") - except in initStats()
-    private final LinkedList<Long> waitTimes = new LinkedList<Long>(); // @GuardedBy("activeTimes") - except in initStats()
-    private final Object maxBorrowWaitTimeMillisLock = new Object();
-    private volatile long maxBorrowWaitTimeMillis = 0; // @GuardedBy("maxBorrowWaitTimeMillisLock")
-    private SwallowedExceptionListener swallowedExceptionListener = null;
+    private final StatsStore activeTimes = new StatsStore(MEAN_TIMING_STATS_CACHE_SIZE);
+    private final StatsStore idleTimes = new StatsStore(MEAN_TIMING_STATS_CACHE_SIZE);
+    private final StatsStore waitTimes = new StatsStore(MEAN_TIMING_STATS_CACHE_SIZE);
+    private final AtomicLong maxBorrowWaitTimeMillis = new AtomicLong(0L);
+    private volatile SwallowedExceptionListener swallowedExceptionListener = null;
 
 
     /**
@@ -138,11 +137,15 @@ public abstract class BaseGenericObjectPool<T> {
         // Populate the creation stack trace
         this.creationStackTrace = getStackTrace(new Exception());
 
-        // save the current CCL to be used later by the evictor Thread
-        factoryClassLoader = Thread.currentThread().getContextClassLoader();
+        // save the current TCCL (if any) to be used later by the evictor Thread
+        ClassLoader cl = Thread.currentThread().getContextClassLoader();
+        if (cl == null) {
+            factoryClassLoader = null;
+        } else {
+            factoryClassLoader = new WeakReference<ClassLoader>(cl);
+        }
 
-        // Initialise the attributes used to record rolling averages
-        initStats();
+        fairness = config.getFairness();
     }
 
 
@@ -256,6 +259,17 @@ public abstract class BaseGenericObjectPool<T> {
         return lifo;
     }
 
+    /**
+     * Returns whether or not the pool serves threads waiting to borrow objects fairly.
+     * True means that waiting threads are served as if waiting in a FIFO queue.
+     *
+     * @return <code>true</code> if waiting threads are to be served
+     *             by the pool in arrival order
+     */
+    public final boolean getFairness() {
+        return fairness;
+    }
+
     /**
      * Sets whether the pool has LIFO (last in, first out) behaviour with
      * respect to idle objects - always returning the most recently used object
@@ -581,7 +595,8 @@ public abstract class BaseGenericObjectPool<T> {
     public final void setEvictionPolicyClassName(
             String evictionPolicyClassName) {
         try {
-            Class<?> clazz = Class.forName(evictionPolicyClassName);
+            Class<?> clazz = Class.forName(evictionPolicyClassName, true,
+                    Thread.currentThread().getContextClassLoader());
             Object policy = clazz.newInstance();
             if (policy instanceof EvictionPolicy<?>) {
                 @SuppressWarnings("unchecked") // safe, because we just checked the class
@@ -767,7 +782,7 @@ public abstract class BaseGenericObjectPool<T> {
      * recently returned objects
      */
     public final long getMeanActiveTimeMillis() {
-        return getMeanFromStatsCache(activeTimes);
+        return activeTimes.getMean();
     }
 
     /**
@@ -777,7 +792,7 @@ public abstract class BaseGenericObjectPool<T> {
      * borrowed objects
      */
     public final long getMeanIdleTimeMillis() {
-        return getMeanFromStatsCache(idleTimes);
+        return idleTimes.getMean();
     }
 
     /**
@@ -787,7 +802,7 @@ public abstract class BaseGenericObjectPool<T> {
      * to wait to borrow an object from the pool
      */
     public final long getMeanBorrowWaitTimeMillis() {
-        return getMeanFromStatsCache(waitTimes);
+        return waitTimes.getMean();
     }
 
     /**
@@ -795,7 +810,7 @@ public abstract class BaseGenericObjectPool<T> {
      * @return maximum wait time in milliseconds since the pool was created
      */
     public final long getMaxBorrowWaitTimeMillis() {
-        return maxBorrowWaitTimeMillis;
+        return maxBorrowWaitTimeMillis.get();
     }
 
     /**
@@ -857,19 +872,17 @@ public abstract class BaseGenericObjectPool<T> {
      */
     final void updateStatsBorrow(PooledObject<T> p, long waitTime) {
         borrowedCount.incrementAndGet();
-        synchronized (idleTimes) {
-            idleTimes.add(Long.valueOf(p.getIdleTimeMillis()));
-            idleTimes.poll();
-        }
-        synchronized (waitTimes) {
-            waitTimes.add(Long.valueOf(waitTime));
-            waitTimes.poll();
-        }
-        synchronized (maxBorrowWaitTimeMillisLock) {
-            if (waitTime > maxBorrowWaitTimeMillis) {
-                maxBorrowWaitTimeMillis = waitTime;
+        idleTimes.add(p.getIdleTimeMillis());
+        waitTimes.add(waitTime);
+
+        // lock-free optimistic-locking maximum
+        long currentMax;
+        do {
+            currentMax = maxBorrowWaitTimeMillis.get();
+            if (currentMax >= waitTime) {
+                break;
             }
-        }
+        } while (!maxBorrowWaitTimeMillis.compareAndSet(currentMax, waitTime));
     }
 
     /**
@@ -879,10 +892,7 @@ public abstract class BaseGenericObjectPool<T> {
      */
     final void updateStatsReturn(long activeTime) {
         returnedCount.incrementAndGet();
-        synchronized (activeTimes) {
-            activeTimes.add(Long.valueOf(activeTime));
-            activeTimes.poll();
-        }
+        activeTimes.add(activeTime);
     }
 
     /**
@@ -977,44 +987,6 @@ public abstract class BaseGenericObjectPool<T> {
         return w.toString();
     }
 
-    /**
-     * Returns the greatest integer less than ore equal to the arithmetic mean
-     * of the entries in <code>cache,</code> acquiring and holding the argument's
-     * monitor while making a local copy.
-     * @param cache list containing entries to analyze
-     * @return truncated arithmetic mean
-     */
-    private long getMeanFromStatsCache(LinkedList<Long> cache) {
-        List<Long> times = new ArrayList<Long>(MEAN_TIMING_STATS_CACHE_SIZE);
-        synchronized (cache) {
-            times.addAll(cache);
-        }
-        double result = 0;
-        int counter = 0;
-        Iterator<Long> iter = times.iterator();
-        while (iter.hasNext()) {
-            Long time = iter.next();
-            if (time != null) {
-                counter++;
-                result = result * ((counter - 1) / (double) counter) +
-                        time.longValue()/(double) counter;
-            }
-        }
-        return (long) result;
-    }
-
-    /**
-     * Initializes pool statistics.
-     */
-    private void initStats() {
-        for (int i = 0; i < MEAN_TIMING_STATS_CACHE_SIZE; i++) {
-            activeTimes.add(null);
-            idleTimes.add(null);
-            waitTimes.add(null);
-        }
-    }
-
-
     // Inner classes
 
     /**
@@ -1036,9 +1008,18 @@ public abstract class BaseGenericObjectPool<T> {
             ClassLoader savedClassLoader =
                     Thread.currentThread().getContextClassLoader();
             try {
-                // Set the class loader for the factory
-                Thread.currentThread().setContextClassLoader(
-                        factoryClassLoader);
+                if (factoryClassLoader != null) {
+                    // Set the class loader for the factory
+                    ClassLoader cl = factoryClassLoader.get();
+                    if (cl == null) {
+                        // The pool has been dereferenced and the class loader
+                        // GC'd. Cancel this timer so the pool can be GC'd as
+                        // well.
+                        cancel();
+                        return;
+                    }
+                    Thread.currentThread().setContextClassLoader(cl);
+                }
 
                 // Evict from the pool
                 try {
@@ -1062,4 +1043,61 @@ public abstract class BaseGenericObjectPool<T> {
             }
         }
     }
+
+    /**
+     * Maintains a cache of values for a single metric and reports
+     * statistics on the cached values.
+     */
+    private class StatsStore {
+
+        private final AtomicLong values[];
+        private final int size;
+        private int index;
+
+        /**
+         * Create a StatsStore with the given cache size.
+         *
+         * @param size number of values to maintain in the cache.
+         */
+        public StatsStore(int size) {
+            this.size = size;
+            values = new AtomicLong[size];
+            for (int i = 0; i < size; i++) {
+                values[i] = new AtomicLong(-1);
+            }
+        }
+
+        /**
+         * Adds a value to the cache.  If the cache is full, one of the
+         * existing values is replaced by the new value.
+         *
+         * @param value new value to add to the cache.
+         */
+        public synchronized void add(long value) {
+            values[index].set(value);
+            index++;
+            if (index == size) {
+                index = 0;
+            }
+        }
+
+        /**
+         * Returns the mean of the cached values.
+         *
+         * @return the mean of the cache, truncated to long
+         */
+        public long getMean() {
+            double result = 0;
+            int counter = 0;
+            for (int i = 0; i < size; i++) {
+                long value = values[i].get();
+                if (value != -1) {
+                    counter++;
+                    result = result * ((counter - 1) / (double) counter) +
+                            value/(double) counter;
+                }
+            }
+            return (long) result;
+        }
+    }
 }
\ No newline at end of file
