@@ -72,7 +72,7 @@ import org.apache.commons.pool2.UsageTracking;
  *
  * @param <T> Type of element pooled in this pool.
  *
- * @version $Revision: 1569016 $
+ * @version $Revision: 1627011 $
  *
  * @since 2.0
  */
@@ -112,6 +112,8 @@ public class GenericObjectPool<T> extends BaseGenericObjectPool<T>
         }
         this.factory = factory;
 
+        idleObjects = new LinkedBlockingDeque<PooledObject<T>>(config.getFairness());
+
         setConfig(config);
 
         startEvictor(getTimeBetweenEvictionRunsMillis());
@@ -423,24 +425,24 @@ public class GenericObjectPool<T> extends BaseGenericObjectPool<T>
         boolean blockWhenExhausted = getBlockWhenExhausted();
 
         boolean create;
-        long waitTime = 0;
+        long waitTime = System.currentTimeMillis();
 
         while (p == null) {
             create = false;
             if (blockWhenExhausted) {
                 p = idleObjects.pollFirst();
                 if (p == null) {
-                    create = true;
                     p = create();
+                    if (p != null) {
+                        create = true;
+                    }
                 }
                 if (p == null) {
                     if (borrowMaxWaitMillis < 0) {
                         p = idleObjects.takeFirst();
                     } else {
-                        waitTime = System.currentTimeMillis();
                         p = idleObjects.pollFirst(borrowMaxWaitMillis,
                                 TimeUnit.MILLISECONDS);
-                        waitTime = System.currentTimeMillis() - waitTime;
                     }
                 }
                 if (p == null) {
@@ -453,8 +455,10 @@ public class GenericObjectPool<T> extends BaseGenericObjectPool<T>
             } else {
                 p = idleObjects.pollFirst();
                 if (p == null) {
-                    create = true;
                     p = create();
+                    if (p != null) {
+                        create = true;
+                    }
                 }
                 if (p == null) {
                     throw new NoSuchElementException("Pool exhausted");
@@ -509,7 +513,7 @@ public class GenericObjectPool<T> extends BaseGenericObjectPool<T>
             }
         }
 
-        updateStatsBorrow(p, waitTime);
+        updateStatsBorrow(p, System.currentTimeMillis() - waitTime);
 
         return p.getObject();
     }
@@ -546,7 +550,7 @@ public class GenericObjectPool<T> extends BaseGenericObjectPool<T>
                     final PooledObjectState state = p.getState();
                     if (state != PooledObjectState.ALLOCATED) {
                         throw new IllegalStateException(
-                                "Object has already been retured to this pool or is invalid");
+                                "Object has already been returned to this pool or is invalid");
                     } else {
                         p.markReturning(); // Keep from being marked abandoned
                     }
@@ -593,7 +597,7 @@ public class GenericObjectPool<T> extends BaseGenericObjectPool<T>
 
         if (!p.deallocate()) {
             throw new IllegalStateException(
-                    "Object has already been retured to this pool or is invalid");
+                    "Object has already been returned to this pool or is invalid");
         }
 
         int maxIdleSave = getMaxIdle();
@@ -609,6 +613,12 @@ public class GenericObjectPool<T> extends BaseGenericObjectPool<T>
             } else {
                 idleObjects.addLast(p);
             }
+            if (isClosed()) {
+                // Pool closed while object was being added to idle objects.
+                // Make sure the returned object is destroyed rather than left
+                // in the idle object pool (which would effectively be a leak)
+                clear();
+            }
         }
         updateStatsReturn(activeTime);
     }
@@ -770,8 +780,23 @@ public class GenericObjectPool<T> extends BaseGenericObjectPool<T>
                         continue;
                     }
 
-                    if (evictionPolicy.evict(evictionConfig, underTest,
-                            idleObjects.size())) {
+                    // User provided eviction policy could throw all sorts of
+                    // crazy exceptions. Protect against such an exception
+                    // killing the eviction thread.
+                    boolean evict;
+                    try {
+                        evict = evictionPolicy.evict(evictionConfig, underTest,
+                                idleObjects.size());
+                    } catch (Throwable t) {
+                        // Slightly convoluted as SwallowedExceptionListener
+                        // uses Exception rather than Throwable
+                        PoolUtils.checkRethrow(t);
+                        swallowException(new Exception(t));
+                        // Don't evict on error conditions
+                        evict = false;
+                    }
+
+                    if (evict) {
                         destroy(underTest);
                         destroyedByEvictorCount.incrementAndGet();
                     } else {
@@ -904,6 +929,12 @@ public class GenericObjectPool<T> extends BaseGenericObjectPool<T>
                 idleObjects.addLast(p);
             }
         }
+        if (isClosed()) {
+            // Pool closed while object was being added to idle objects.
+            // Make sure the returned object is destroyed rather than left
+            // in the idle object pool (which would effectively be a leak)
+            clear();
+        }
     }
 
     /**
@@ -1098,8 +1129,7 @@ public class GenericObjectPool<T> extends BaseGenericObjectPool<T>
      * {@link #_maxActive} objects created at any one time.
      */
     private final AtomicLong createCount = new AtomicLong(0);
-    private final LinkedBlockingDeque<PooledObject<T>> idleObjects =
-        new LinkedBlockingDeque<PooledObject<T>>();
+    private final LinkedBlockingDeque<PooledObject<T>> idleObjects;
 
     // JMX specific attributes
     private static final String ONAME_BASE =
