@@ -32,6 +32,8 @@ import java.util.ArrayList;
 import java.util.NoSuchElementException;
 import java.util.Random;
 import java.util.Set;
+import java.util.Timer;
+import java.util.TimerTask;
 import java.util.concurrent.Callable;
 import java.util.concurrent.ConcurrentHashMap;
 import java.util.concurrent.ExecutorService;
@@ -60,7 +62,7 @@ import org.junit.Before;
 import org.junit.Test;
 
 /**
- * @version $Revision: 1566584 $
+ * @version $Revision: 1648001 $
  */
 public class TestGenericKeyedObjectPool extends TestKeyedObjectPool {
 
@@ -413,6 +415,28 @@ public class TestGenericKeyedObjectPool extends TestKeyedObjectPool {
             assertFalse(pool.getBlockWhenExhausted());
         }
     }
+    
+    /** 
+     * JIRA: POOL-270 - make sure constructor correctly sets run
+     * frequency of evictor timer.
+     */
+    @Test
+    public void testContructorEvictionConfig() throws Exception {
+        GenericKeyedObjectPoolConfig config = new GenericKeyedObjectPoolConfig();
+        config.setTimeBetweenEvictionRunsMillis(500);
+        config.setMinEvictableIdleTimeMillis(50);
+        config.setNumTestsPerEvictionRun(5);
+        GenericKeyedObjectPool<String, String> p = new GenericKeyedObjectPool<String, String>(factory, config);
+        for(int i=0;i<5;i++) {
+            p.addObject("one");
+        }
+        try { Thread.sleep(100); } catch(InterruptedException e) { }
+        assertEquals(5, p.getNumIdle("one"));
+        try { Thread.sleep(500); } catch(InterruptedException e) { }
+        assertEquals(0, p.getNumIdle("one"));
+        p.close();
+    }
+    
 
     @Test(timeout=60000)
     public void testEviction() throws Exception {
@@ -1030,6 +1054,68 @@ public class TestGenericKeyedObjectPool extends TestKeyedObjectPool {
         }
     }
 
+    /*
+     * Note: This test relies on timing for correct execution. There *should* be
+     * enough margin for this to work correctly on most (all?) systems but be
+     * aware of this if you see a failure of this test.
+     */
+    @SuppressWarnings({
+        "rawtypes", "unchecked"
+    })
+    @Test(timeout=60000)
+    public void testBorrowObjectFairness() throws Exception {
+
+        int numThreads = 40;
+        int maxTotal = 40;
+
+        GenericKeyedObjectPoolConfig config = new GenericKeyedObjectPoolConfig();
+        config.setMaxTotalPerKey(maxTotal);
+        config.setFairness(true);
+        config.setLifo(false);
+        config.setMaxIdlePerKey(maxTotal);
+
+        pool = new GenericKeyedObjectPool<String, String>(factory, config);
+
+        // Exhaust the pool
+        String[] objects = new String[maxTotal];
+        for (int i = 0; i < maxTotal; i++) {
+            objects[i] = pool.borrowObject("0");
+        }
+
+        // Start and park threads waiting to borrow objects
+        TestThread[] threads = new TestThread[numThreads];
+        for(int i=0;i<numThreads;i++) {
+            threads[i] = new TestThread(pool, 1, 0, 2000, false, "0" + String.valueOf(i % maxTotal), "0");
+            Thread t = new Thread(threads[i]);
+            t.start();
+            // Short delay to ensure threads start in correct order
+            try {
+                Thread.sleep(10);
+            } catch (InterruptedException e) {
+                fail(e.toString());
+            }
+        }
+
+        // Return objects, other threads should get served in order
+        for (int i = 0; i < maxTotal; i++) {
+            pool.returnObject("0", objects[i]);
+        }
+
+        // Wait for threads to finish
+        for(int i=0;i<numThreads;i++) {
+            while(!(threads[i]).complete()) {
+                try {
+                    Thread.sleep(500L);
+                } catch(InterruptedException e) {
+                    // ignored
+                }
+            }
+            if(threads[i].failed()) {
+                fail("Thread "+i+" failed: "+threads[i]._exception.toString());
+            }
+        }
+    }
+
     @Test(timeout=60000)
     public void testConstructors() throws Exception {
 
@@ -1250,13 +1336,13 @@ public class TestGenericKeyedObjectPool extends TestKeyedObjectPool {
             wtt[i] = new WaitingTestThread(pool,Integer.toString(i % keyCount),holdTime);
         }
         long origin = System.currentTimeMillis()-1000;
-        for(int i=0; i < wtt.length; i++){
-            wtt[i].start();
+        for (WaitingTestThread element : wtt) {
+            element.start();
         }
         int failed = 0;
-        for(int i=0; i < wtt.length; i++){
-            wtt[i].join();
-            if (wtt[i]._thrown != null){
+        for (WaitingTestThread element : wtt) {
+            element.join();
+            if (element._thrown != null){
                 failed++;
             }
         }
@@ -1269,8 +1355,7 @@ public class TestGenericKeyedObjectPool extends TestKeyedObjectPool {
                     " Threads: " + wtt.length +
                     " Failed: " + failed
                     );
-            for(int i=0; i < wtt.length; i++){
-                WaitingTestThread wt = wtt[i];
+            for (WaitingTestThread wt : wtt) {
                 System.out.println(
                         "Preborrow: " + (wt.preborrow-origin) +
                         " Postborrow: " + (wt.postborrow != 0 ? wt.postborrow-origin : -1) +
@@ -1485,6 +1570,55 @@ public class TestGenericKeyedObjectPool extends TestKeyedObjectPool {
         Assert.assertEquals(nIterations, pool.getDestroyedCount());
     }
 
+    // POOL-259
+    @Test
+    public void testClientWaitStats() throws Exception {
+        SimpleFactory<String> factory = new SimpleFactory<String>();
+        // Give makeObject a little latency
+        factory.setMakeLatency(200);
+        final GenericKeyedObjectPool<String, String> pool = new GenericKeyedObjectPool<String, String>(
+                factory, new GenericKeyedObjectPoolConfig());
+        String s = pool.borrowObject("one");
+        // First borrow waits on create, so wait time should be at least 200 ms
+        // Allow 100ms error in clock times
+        Assert.assertTrue(pool.getMaxBorrowWaitTimeMillis() >= 100);
+        Assert.assertTrue(pool.getMeanBorrowWaitTimeMillis() >= 100);
+        pool.returnObject("one", s);
+        pool.borrowObject("one");
+        // Second borrow does not have to wait on create, average should be about 100
+        Assert.assertTrue(pool.getMaxBorrowWaitTimeMillis() > 100);
+        Assert.assertTrue(pool.getMeanBorrowWaitTimeMillis() < 200);
+        Assert.assertTrue(pool.getMeanBorrowWaitTimeMillis() > 20);
+    }
+
+    // POOL-276
+    @Test
+    public void testValidationOnCreateOnly() throws Exception {
+        factory.enableValidation = true;
+
+        pool.setMaxTotal(1);
+        pool.setTestOnCreate(true);
+        pool.setTestOnBorrow(false);
+        pool.setTestOnReturn(false);
+        pool.setTestWhileIdle(false);
+
+        final String o1 = pool.borrowObject("KEY");
+        Assert.assertEquals("KEY0", o1);
+        Timer t = new Timer();
+        t.schedule(
+                new TimerTask() {
+                    @Override
+                    public void run() {
+                        pool.returnObject("KEY", o1);
+                    }
+                }, 3000);
+
+        String o2 = pool.borrowObject("KEY");
+        Assert.assertEquals("KEY0", o2);
+
+        Assert.assertEquals(1, factory.validateCounter);
+    }
+
     /**
      * Attempts to invalidate an object, swallowing IllegalStateException.
      */
@@ -1583,27 +1717,47 @@ public class TestGenericKeyedObjectPool extends TestKeyedObjectPool {
     static class TestThread<T> implements Runnable {
         private final java.util.Random _random = new java.util.Random();
 
-        // Thread config items
+        /** GKOP to hit */
         private final KeyedObjectPool<String,T> _pool;
+        /** number of borrow/return iterations */
         private final int _iter;
-        private final int _delay;
+        /** delay before borrow */
+        private final int _startDelay;
+        /** delay before return */
+        private final int _holdTime;
+        /** whether or not delays are random (with max = configured values) */
+        private final boolean _randomDelay;
+        /** expected object */
+        private final T _expectedObject;
+        /** key used in borrow / return sequence - null means random */
+        private final String _key;
 
         private volatile boolean _complete = false;
         private volatile boolean _failed = false;
         private volatile Exception _exception;
 
         public TestThread(KeyedObjectPool<String,T> pool) {
-            this(pool, 100, 50);
+            this(pool, 100, 50, 50, true, null, null);
         }
 
         public TestThread(KeyedObjectPool<String,T> pool, int iter) {
-            this(pool, iter, 50);
+            this(pool, iter, 50, 50, true, null, null);
         }
 
         public TestThread(KeyedObjectPool<String,T> pool, int iter, int delay) {
+            this(pool, iter, delay, delay, true, null, null);
+        }
+
+        public TestThread(KeyedObjectPool<String,T> pool, int iter, int startDelay,
+            int holdTime, boolean randomDelay, T expectedObject, String key) {
             _pool = pool;
             _iter = iter;
-            _delay = delay;
+            _startDelay = startDelay;
+            _holdTime = holdTime;
+            _randomDelay = randomDelay;
+            _expectedObject = expectedObject;
+            _key = key;
+
         }
 
         public boolean complete() {
@@ -1617,9 +1771,9 @@ public class TestGenericKeyedObjectPool extends TestKeyedObjectPool {
         @Override
         public void run() {
             for(int i=0;i<_iter;i++) {
-                String key = String.valueOf(_random.nextInt(3));
+                String key = _key == null ? String.valueOf(_random.nextInt(3)) : _key;
                 try {
-                    Thread.sleep(_random.nextInt(_delay));
+                    Thread.sleep(_randomDelay ? _random.nextInt(_startDelay) : _startDelay);
                 } catch(InterruptedException e) {
                     // ignored
                 }
@@ -1633,8 +1787,15 @@ public class TestGenericKeyedObjectPool extends TestKeyedObjectPool {
                     break;
                 }
 
+                if (_expectedObject != null && !_expectedObject.equals(obj)) {
+                    _exception = new Exception("Expected: "+_expectedObject+ " found: "+obj);
+                    _failed = true;
+                    _complete = true;
+                    break;
+                }
+
                 try {
-                    Thread.sleep(_random.nextInt(_delay));
+                    Thread.sleep(_randomDelay ? _random.nextInt(_holdTime) : _holdTime);
                 } catch(InterruptedException e) {
                     // ignored
                 }
@@ -1663,6 +1824,7 @@ public class TestGenericKeyedObjectPool extends TestKeyedObjectPool {
             if (exceptionOnCreate) {
                 throw new Exception();
             }
+            doWait(makeLatency);
             String out = null;
             synchronized(this) {
                 activeCount++;
@@ -1713,6 +1875,9 @@ public class TestGenericKeyedObjectPool extends TestKeyedObjectPool {
         public void setDestroyLatency(long destroyLatency) {
             this.destroyLatency = destroyLatency;
         }
+        public void setMakeLatency(long makeLatency) {
+            this.makeLatency = makeLatency;
+        }
         public void setValidationEnabled(boolean b) {
             enableValidation = b;
         }
@@ -1745,6 +1910,7 @@ public class TestGenericKeyedObjectPool extends TestKeyedObjectPool {
         boolean oddValid = true;
         boolean enableValidation = false;
         long destroyLatency = 0;
+        long makeLatency = 0;
         volatile int maxTotalPerKey = Integer.MAX_VALUE;
         boolean exceptionOnPassivate = false;
         boolean exceptionOnActivate = false;
